local shared = require "shared"
local spawnore = require "prototypes.spawnore"
gui = {}
chunkManager = require "prototypes.chunk_limit"
gui = require "prototypes.gui"

local script_data = {
  members = 0,
  donations = {
    amount = 0,
    bought = {
      ["-2"] = {
        ["-2"] = true,
        ["-1"] = true,
        ["0"] = true,
        ["1"] = true,
        ["2"] = true
      },
      ["-1"] = {
        ["-2"] = true,
        ["-1"] = true,
        ["0"] = true,
        ["1"] = true,
        ["2"] = true
      },
      ["0"] = {
        ["-2"] = true,
        ["-1"] = true,
        ["0"] = true,
        ["1"] = true,
        ["2"] = true
      },
      ["1"] = {
        ["-2"] = true,
        ["-1"] = true,
        ["0"] = true,
        ["1"] = true,
        ["2"] = true
      },
      ["2"] = {
        ["-2"] = true,
        ["-1"] = true,
        ["0"] = true,
        ["1"] = true,
        ["2"] = true
      }
    },
    current_target = {}
  }
}

global.chunkRegister = {}
chunkRegister = global.chunkRegister
meta_stringkeys = {
  __newindex = function(ta, key, val)
    rawset(ta, tostring(key), val)
  end,
  __index = function(ta, key)
    if type(key) ~= "string" and key ~= nil then
      return ta[tostring(key)]
    end
  end
}

setmetatable(chunkRegister, meta_stringkeys)
setmetatable(script_data.donations.bought, meta_stringkeys)
-- local config = require("__stdlib__/stdlib/config")
-- config.skip_script_protections = true

Core = require("__stdlib__/stdlib/core")
Game = require("__stdlib__/stdlib/game")
Event = require("__stdlib__/stdlib/event/event")
Gui = require("__stdlib__/stdlib/event/gui")
Position = require("__stdlib__/stdlib/area/position")
Surface = require("__stdlib__/stdlib/area/surface")
Player = require("__stdlib__/stdlib/event/player")
table = require("__stdlib__/stdlib/utils/table")

chunkManager.setup_events()
spawnore.setup_events()
gui.init_events()

local register_api_events = function()
  events = remote.call("tvc_api", "get_events")

  -- Ensure no duplicate event registration
  Event.remove {
    events.api_on_donation,
    events.api_on_member,
    events.api_on_raid,
    events.api_on_host,
    events.api_on_channel_point
  }

  Event.register(
    events.api_on_donation,
    function(event)
      local message = event.message
      if not message.isTest or (message.isTest and global.allow_test) then
        local amount = tonumber(message.amount)
        if message.type == "bits" then
          amount = amount / 100
        end

        script_data.donations.amount = script_data.donations.amount + amount
      else
        game.print("Found test message, ignoring")
      end

      if message.type == "bits" then
        game.print({"tvc-message-donation", message.name, message.amount .. " Bits", message.message})
      else
        local amount = message.formatted_amount or message.amount .. message.currency
        game.print({"tvc-message-donation", message.name, amount, message.message})
      end
    end
  )

  Event.register(
    events.api_on_member,
    function(event)
      local message = event.message
      local do_action = false

      if not message.isTest or (message.isTest and global.allow_test) then
        local name = message.display_name or message.name
        if message["for"] == "youtube_account" then
          message.sub_plan = "1000"
        end
        if string.lower(message.sub_plan) == "prime" then
          message.sub_plan = "1000"
        end
        do_action = true
      else
        game.print("Found test message, ignoring")
      end

      if (message["streak-months"] and message["streak-months"] > 1) or message.sub_type == "resub" then
        game.print({"tvc-message-resub", message.name, message.months})
      elseif message.gifter_display_name or message.gifter then
        game.print({"tvc-message-member-gift", message.name, message.gifter_display_name or message.gifter})
      else
        local name = message.display_name or message.name
        if message.type == "pledge" then
          name = "A new patreon"
        end
        game.print({"tvc-message-member", name})
      end

      if do_action then
        script_data.members = script_data.members + 1

        if script_data.members % settings.global[shared.settings.member_spawn_ore].value == 0 then
          spawnore.find_spawn_area_in_visible_area()
        end
      end
    end
  )

  Event.register(
    events.api_on_raid,
    function(event)
      local message = event.message
      local min_amount = settings.global[shared.settings.min_raid_host].value
      if message.raiders and tonumber(message.raiders) >= min_amount then
        game.print({"tvc-raid", message.name, message.raiders})
      end
    end
  )

  Event.register(
    events.api_on_host,
    function(event)
      local message = event.message
      local min_amount = settings.global[shared.settings.min_raid_host].value
      if message.viewers and tonumber(message.viewers) >= min_amount then
        game.print({"tvc-host", message.name, message.viewers})
      end
    end
  )

  Event.register(
    events.api_on_channel_point,
    function(event)
      local message = event.message
      local name = message.display_name or message.name
      if not message.isTest or (message.isTest and global.allow_test) then
      else
        game.print("Found test message, ignoring")
      end
    end
  )
end

Event.register(
  defines.events.on_tick,
  function(event)
    if event.tick == 120 then
      local surface = game.surfaces[1]
      for x, row in pairs(script_data.donations.bought) do
        for y, _ in pairs(row) do
          x = tonumber(x)
          y = tonumber(y)
          chunkManager.buy {x = x, y = y}

          local ents =
            surface.find_entities_filtered {
            name = {shared.entity.chunk_entity, shared.entity.chunk_border},
            area = {
              {x * 32 + 4, y * 32 + 4},
              {x * 32 + 28, y * 32 + 28}
            }
          }
          for _, ent in pairs(ents) do
            ent.destroy()
          end
        end
      end
    end
  end
)

Event.on_init(
  function(event)
    global.control_data = global.control_data or script_data
    register_api_events()
  end
)

Event.on_load(
  function(event)
    script_data = global.control_data or script_data
    register_api_events()

    chunkRegister = global.chunkRegister
    setmetatable(chunkRegister, meta_stringkeys)
    for _, v in pairs(chunkRegister) do
      setmetatable(v, meta_stringkeys)
      if chunkRegister[v] ~= nil then
        for _, x in pairs(chunkRegister[v]) do
          setmetatable(x, meta_stringkeys)
        end
      end
    end
    local bought = script_data.donations.bought
    setmetatable(bought, meta_stringkeys)
    for _, v in pairs(bought) do
      setmetatable(v, meta_stringkeys)
      if bought[v] ~= nil then
        for _, x in pairs(bought[v]) do
          setmetatable(x, meta_stringkeys)
        end
      end
    end
  end
)

remote.add_interface(
  "nilaus",
  {
    allow_test = function(bool)
      global.allow_test = bool
    end,
    test_spawn = function(teleport)
      local loc = spawnore.find_spawn_area_in_visible_area()
      if teleport and loc then
        game.players[1].teleport(loc)
      end
    end,
    buy = function(x, y)
      chunkManager.buy({x = x, y = y})
    end,
    regen_entities = function(chunks)
      chunks = chunks or 50
      local force = game.forces.player

      for _, surface in pairs(game.surfaces) do
        for _, ent in pairs(
          surface.find_entities_filtered(
            {name = shared.chunk_entity, area = {{chunks * 32 * -1, chunks * 32 * -1}, {chunks * 32, chunks * 32}}}
          )
        ) do
          if (ent.valid) then
            local new_ent =
              surface.create_entity({name = shared.chunk_entity, position = ent.position, force = ent.force})
            if (force.is_chunk_charted(surface, Chunk.from_position(ent.position))) then
              force.chart(surface, Position(ent.position):expand_to_area(32))
            end
            new_ent.destructible = false
            new_ent.minable = true
            ent.destroy()
          end
        end
      end
    end,
    recheck_chunks = function(size)
      local from = 0 - (size / 2)
      local to = size / 2
      local surface = game.surfaces.nauvis

      for x = from, to do
        for y = from, to do
          if not chunkManager.isBought({x = x, y = y}) then
            local area = {left_top = {x = x * 32, y = y * 32}, right_bottom = {x = x * 32 + 32, y = y * 32 + 32}}
            local ents = surface.find_entities_filtered({name = shared.chunk_entity, area = area})
            if #ents == 0 then
              local new_ent =
                surface.create_entity(
                {
                  name = shared.chunk_entity,
                  position = {x = area.left_top.x + 16, y = area.left_top.y + 16},
                  force = game.forces.player
                }
              )
              new_ent.destructible = false
              new_ent.minable = true
            elseif ents[1].destructible == true then
              ents[1].destructible = false
              ents[1].minable = true
            end
          end
        end
      end
    end,
    remove_lines = function()
      for x, row in pairs(chunkRegister) do
        for y, v in pairs(row) do
          if v.lines then
            for _, b in pairs({"top", "left", "right", "bottom"}) do
              if v.lines[b] ~= nil and type(v.lines[b]) == "number" then
                rendering.destroy(v.lines[b])
              end
            end
            v.rendered = false
            v.lines = nil
          end
        end
      end

      chunkManager.recheckChunks()
    end
  }
)
