local spawnore = {}
local spawn_ore_order = {
  "iron-ore",
  "copper-ore",
  "crude-oil",
  "coal",
  "stone"
}

local shared = require "shared"

local script_data = {
  last_spawn = #spawn_ore_order,
  spawn_new_on_tick = false,
  round = 1,
  chunkRegister = {}
}

meta_stringkeys = {
  __newindex = function(ta, key, val)
    rawset(ta, tostring(key), val)
  end,
  __index = function(ta, key)
    if type(key) ~= "string" and key ~= nil then
      return ta[tostring(key)]
    end
  end
}

setmetatable(script_data.chunkRegister, meta_stringkeys)

function spawnore.setup_events()
  Event.on_load(
    function(event)
      script_data = global.spawnore_data or script_data
      chunkRegister = script_data.chunkRegister
      setmetatable(chunkRegister, meta_stringkeys)
      for _, v in pairs(chunkRegister) do
        setmetatable(v, meta_stringkeys)
        if chunkRegister[v] ~= nil then
          for _, x in pairs(chunkRegister[v]) do
            setmetatable(x, meta_stringkeys)
          end
        end
      end
    end
  )

  Event.on_init(
    function(event)
      global.spawnore_data = global.spawnore_data or script_data
    end
  )

  Event.register(defines.events.on_chunk_charted, spawnore.on_charted)
  Event.register(
    -240,
    function()
      if script_data.spawn_new_on_tick then
        script_data.spawn_new_on_tick = false
        spawnore.find_spawn_area_in_visible_area()
      end
    end
  )
end

function spawnore.on_charted(event)
  if (script_data.chunkRegister[event.position.x] == nil) then
    script_data.chunkRegister[event.position.x] = {}
    setmetatable(script_data.chunkRegister[event.position.x], meta_stringkeys)
  end
  if (script_data.chunkRegister[event.position.x][event.position.y] == nil) then
    script_data.chunkRegister[event.position.x][event.position.y] = {
      visible = true
    }
  end
end

local function create_patch(surface, ore_name, position, size)
  local size = size or 5
  local richness = settings.global[shared.settings.spawn_patch_size].value * math.pow(10, script_data.round)
  local landfill = {}
  local rand = math.random
  if game.entity_prototypes[ore_name].infinite_resource then
    size = math.floor(size / 3)
    for y = (-size + 1), size do
      for x = (-size + 1), size do
        for i = -1, 1 do
          for j = -1, 1 do
            if not surface.get_tile(position.x + (x * 4) + i, position.y + (y * 4) + j).collides_with("ground-tile") then
              table.insert(
                landfill,
                {name = "landfill", position = {position.x + (x * 4) + i, position.y + (y * 4) + j}}
              )
            end
          end
        end

        -- randomize richeness
        local cur_rich = richness * 200
        if rand(0, 1) == 1 then
          cur_rich = cur_rich - (cur_rich * (rand(0, 70) / 10) / 100)
        else
          cur_rich = cur_rich + (cur_rich * (rand(0, 70) / 10) / 100)
        end

        surface.create_entity(
          {name = ore_name, amount = cur_rich, position = {position.x + (x * 4), position.y + (y * 4)}}
        )
      end
    end
  else
    for y = (-size + 1), size do
      for x = (-size + 1), size do
        if not surface.get_tile(position.x + x, position.y + y).collides_with("ground-tile") then
          table.insert(landfill, {name = "landfill", position = {position.x + x, position.y + y}})
        end

        -- randomize richeness
        local cur_rich = richness
        if rand(0, 1) == 1 then
          cur_rich = cur_rich - (cur_rich * (rand(0, 70) / 10) / 100)
        else
          cur_rich = cur_rich + (cur_rich * (rand(0, 70) / 10) / 100)
        end

        surface.create_entity({name = ore_name, amount = cur_rich, position = {position.x + x, position.y + y}})
      end
    end
  end

  if #landfill > 0 then
    surface.set_tiles(landfill)
  end
end

function spawnore.find_spawn_area_in_visible_area()
  local force = game.forces.player
  local surface = game.surfaces.nauvis
  local random = math.random
  local patch_size = 5

  local end_result = false
  local x = table.keys(script_data.chunkRegister)
  local rounds = 0

  while end_result == false do
    rounds = rounds + 1
    if rounds > 30 then
      script_data.spawn_new_on_tick = true
      break
    end
    local rand_x = tonumber(x[random(1, #x)])
    local y = table.keys(script_data.chunkRegister[rand_x])
    local rand_y = tonumber(y[random(1, #y)])

    if (rand_x < -1 or rand_x > 1) and (rand_y < -1 or rand_y > 1) then
      if
        script_data.chunkRegister[rand_x][rand_y].visible and force.is_chunk_visible(surface, {rand_x, rand_y}) and
          surface.count_entities_filtered(
            {type = "resource", position = {rand_x * 32 + 16, rand_y * 32 + 16}, radius = 32}
          ) == 0
       then
        local pos_x = random(0, 32)
        local pos_y = random(0, 32)
        area = {
          left_top = {(rand_x * 32) + pos_x - patch_size, (rand_y * 32) + pos_y - patch_size},
          right_bottom = {(rand_x * 32) + pos_x + patch_size, (rand_y * 32) + pos_y + patch_size}
        }

        local tiles =
          surface.count_tiles_filtered {
          area = area,
          collision_mask = "ground-tile"
        }

        if tiles > patch_size * 4 then
          end_result = {x = rand_x * 32 + pos_x, y = rand_y * 32 + pos_y}
        end
      end
    end
  end

  if end_result then
    script_data.last_spawn = script_data.last_spawn + 1
    if script_data.last_spawn > #spawn_ore_order then
      script_data.last_spawn = 1
      script_data.round = script_data.round + 1
    end

    create_patch(surface, spawn_ore_order[script_data.last_spawn], end_result, patch_size)
    game.print({"nilaus.ping_gps", end_result.x, end_result.y})
  end

  return end_result
end

return spawnore
